/* verzeichnis.cc
 * This file belongs to Worker, a filemanager for UNIX/X11.
 * Copyright (C) 2001-2007 Ralf Hoffmann.
 * You can contact me at: ralf@boomerangsworld.de
 *   or http://www.boomerangsworld.de/worker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "verzeichnis.hh"
#include "datei.h"
#include <aguix/lowlevelfunc.h>
#include "wconfig.h"
#include "grouphash.h"
#include "worker.h"
#include "wcfiletype.hh"
#include <algorithm>

Verzeichnis::Verzeichnis() : _highest_ID_plus_one( 0 ),
                             _id_serial( -1 )
{
  files=NULL;
  actdir=NULL;
  invalid_entries=NULL;
}

Verzeichnis::~Verzeichnis()
{
  closeDir();
}

ArrayList *Verzeichnis::getFiles()
{
  return files;
}

int Verzeichnis::readDir( const char *dirstr, bool dontWarnInvalid )
{
  if(dirstr==NULL) return -1;
  if(strlen(dirstr)<1) return -1;
  if(files!=NULL) closeDir();
  files=new ArrayList();
  invalid_entries=new ArrayList();
  // nahezu 1 zu 1 vom Original bernehmen

  int returnvalue=0;
  worker_struct_dirent *namelist;
  DIR *dir;
  FileEntry *fe1=NULL;
  char *tstr;
  long x,ldnr;
  FileEntry *fpp;
  long havepointpoint;
  int id, erg;
  bool user_abort = false;
  const char *myerror;
  char *tstr2;
  std::string str1;
  
  fe1=NULL;
  dir = worker_opendir( dirstr );
  if(dir!=NULL) {
//    worker_readdir( dir );  /* berspringt den Eintrag "." */
    havepointpoint=-1;
    fpp=NULL;
    ldnr=0;
    while ( ( namelist = worker_readdir( dir ) ) != NULL ) {
      // ignore "." entries (no need)
      if(strcmp(namelist->d_name,".")!=0) {
          
        if ( strcmp( namelist->d_name, ".." ) == 0 &&
             havepointpoint >= 0 ) {
            // skip multiple .. entries
            continue;
        }
        
        fe1=new FileEntry();
        if(strcmp(namelist->d_name,"..")==0) {
          // this is because of problems with NTFS
          havepointpoint=ldnr;
          fpp=fe1;
        }
        fe1->name=dupstring(namelist->d_name);
        x=strlen(fe1->name);
        tstr=(char*)_allocsafe(strlen(dirstr)+1+x+1);
        strcpy(tstr,dirstr);
        if(strlen(dirstr)>1) strcat(tstr,"/");
        strcat(tstr,namelist->d_name);
        fe1->fullname=tstr;
        ldnr++;
        if(fe1->readInfos()==0) {
          if(fe1->name[0]=='.') fe1->isHidden=true;
          else fe1->isHidden=false;
          files->addElement(fe1);
        } else {
          ldnr--;
          myerror = strerror( errno );
          if ( ( Worker::getRequester() != NULL ) && ( dontWarnInvalid == false ) ) {
            tstr2 = (char*)_allocsafe( strlen( catalog.getLocale( 531 ) ) +
                                       strlen( fe1->fullname ) +
                                       strlen( myerror ) + 1 );
            sprintf( tstr2, catalog.getLocale( 531 ), fe1->fullname, myerror );
            str1 = catalog.getLocale( 11 );
            str1 += "|";
            str1 += catalog.getLocale( 8 );
            erg = Worker::getRequester()->request( catalog.getLocale( 347 ), tstr2, str1.c_str() );
            _freesafe( tstr2 );
            if ( erg == 1 ) {
              user_abort = true;
            }
          }

          if ( strcmp( namelist->d_name, ".." ) == 0 ) {
            // we need a .. entry so reset flag
            havepointpoint = -1;
            fpp = NULL;
            delete fe1;
          } else {
            invalid_entries->addElement(fe1);
          }
        }
      }
      if ( user_abort == true ) break;
    }
    worker_closedir( dir );
    if ( user_abort == false ) {
      if(havepointpoint==-1) {
        // there is no parent-dir-entry
        // create one

        fe1=new FileEntry();
        fe1->name=dupstring("..");
        x=strlen(fe1->name);
        tstr=(char*)_allocsafe(strlen(dirstr)+1+x+1);
        strcpy(tstr,dirstr);
        if(strlen(dirstr)>1) strcat(tstr,"/");
        strcat(tstr,fe1->name);
        fe1->fullname=tstr;

        // try to get some infos about ..
        // it's very likely that this works even when the dir have no ..
        if ( fe1->readInfos() != 0 ) {
          // failed, use some default values
          //TODO: Besonders toll gefllt mir das nicht
          //      Fge ich in FileEntry ein Feld hinzu, muss ich dies u.U. auch hier
          //      init.
          //      Besser wre, wenn ich statbuf direkt in Klasse aufnehme und
          //      ich dann einfach ein memset machen kann
          memset( &( fe1->statbuf ), 0, sizeof ( fe1->statbuf ) );
          memset( &( fe1->dstatbuf ), 0, sizeof ( fe1->dstatbuf ) );
          fe1->statbuf.mode = S_IFDIR | S_IRUSR | S_IWUSR | S_IXUSR;
          fe1->statbuf.userid = (uid_t)-1;
          fe1->statbuf.groupid = (gid_t)-1;
          fe1->isCorrupt = false;
          fe1->isLink = false;
        }

        if(fe1->name[0]=='.') fe1->isHidden=true;
        else fe1->isHidden=false;
        files->addElementAt(0,fe1);
      } else if(havepointpoint>0) {
        if(fpp!=NULL) {
          havepointpoint=files->getIndex(fpp);
          files->removeElementAt(havepointpoint);
          files->addElementAt(0,fpp);
        }
      }
      actdir=dupstring(dirstr);
    }
  } else {
    // requester: can't read directory
    myerror = strerror( errno );
    tstr2 = (char*)_allocsafe( strlen( catalog.getLocale( 532 ) ) +
                               strlen( dirstr ) +
                               strlen( myerror ) + 1 );
    sprintf( tstr2, catalog.getLocale( 532 ), dirstr, myerror );
    str1 = catalog.getLocale( 11 );
    Worker::getRequester()->request( catalog.getLocale( 347 ), tstr2, str1.c_str() );
    _freesafe( tstr2 );
    returnvalue = -1;
    delete files;
    files=NULL;
    delete invalid_entries;
    invalid_entries=NULL;
  }
  if ( user_abort == true ) {
    id=files->initEnum();
    fe1=(FileEntry*)files->getFirstElement(id);
    while(fe1!=NULL) {
      delete fe1;
      fe1=(FileEntry*)files->getNextElement(id);
    }
    files->closeEnum(id);

    id=invalid_entries->initEnum();
    fe1=(FileEntry*)invalid_entries->getFirstElement(id);
    while(fe1!=NULL) {
      delete fe1;
      fe1=(FileEntry*)invalid_entries->getNextElement(id);
    }
    invalid_entries->closeEnum(id);

    returnvalue=1;
    delete files;
    files=NULL;
    delete invalid_entries;
    invalid_entries=NULL;
  }

  createIDs();
  buildID2Pos();
  return returnvalue;
}

int Verzeichnis::sortfunction(void *p1,void *p2,int mode)
{
  FileEntry *fe1,*fe2;
  int comp=0;
  fe1=(FileEntry*)p1;
  fe2=(FileEntry*)p2;
  loff_t s1,s2;
  int dirm=mode&0x600;
  WCFiletype *ft1, *ft2;
  
/*  if(strcmp(fe1->name,"..")==0) return true;  // Sonderfall
  if(strcmp(fe2->name,"..")==0) return false;*/
  if((strcmp(fe1->name,"..")==0)&&(strcmp(fe2->name,"..")==0)) return 0;
  else if(strcmp(fe1->name,"..")==0) return -1;
  else if(strcmp(fe2->name,"..")==0) return 1;

  switch(mode&0xff) {
    case SORT_NAME:
      comp=strcmp(fe1->name,fe2->name);
      break;
    case SORT_SIZE:
      s1=-1;
      s2=-1;
      if ( S_ISCHR( fe1->mode() ) ||
           S_ISBLK( fe1->mode() ) ) {
          s1 = (loff_t)fe1->rdev();
      } else {
        if ( fe1->isDir() == true) s1 = fe1->dirsize;
        if ( s1 < 0 ) s1 = fe1->size();
      }
      if ( S_ISCHR( fe2->mode() ) ||
           S_ISBLK( fe2->mode() ) ) {
          s2 = (loff_t)fe2->rdev();
      } else {
        if ( fe2->isDir() == true ) s2 = fe2->dirsize;
        if ( s2 < 0 ) s2 = fe2->size();
      }
      if(s1<s2) comp=-1;
      else if(s1>s2) comp=1;
      else comp=strcmp(fe1->name,fe2->name);
      break;
    case SORT_ACCTIME:
      if ( fe1->lastaccess() < fe2->lastaccess() ) comp = -1;
      else if ( fe1->lastaccess() > fe2->lastaccess() ) comp = 1;
      else comp=strcmp(fe1->name,fe2->name);
      break;
    case SORT_MODTIME:
      if ( fe1->lastmod() < fe2->lastmod() ) comp = -1;
      else if ( fe1->lastmod() > fe2->lastmod() ) comp = 1;
      else comp=strcmp(fe1->name,fe2->name);
      break;
    case SORT_CHGTIME:
      if ( fe1->lastchange() < fe2->lastchange() ) comp = -1;
      else if ( fe1->lastchange() > fe2->lastchange() ) comp = 1;
      else comp=strcmp(fe1->name,fe2->name);
      break;
    case SORT_TYPE:
      if ( fe1->filetype != NULL ) {
        ft1 = fe1->filetype;
      } else {
        if(fe1->isDir()==true) {
          ft1 = wconfig->getdirtype();
        } else {
          ft1 = wconfig->getnotyettype();
        }
      }
      if ( fe2->filetype != NULL ) {
        ft2 = fe2->filetype;
      } else {
        if(fe2->isDir()==true) {
          ft2 = wconfig->getdirtype();
        } else {
          ft2 = wconfig->getnotyettype();
        }
      }
      comp = strcmp( ft1->getName(), ft2->getName() );
      if ( comp == 0 )
        comp = strcmp( fe1->name, fe2->name );
      break;
    case SORT_OWNER:
      // first cmp usernames
      // if equal cmp groupnames
      // if equal cmp names
      
      // getUserNameS only look for existing entry in hashtable and returns
      // static pointer
      // so it's fast
      //TODO: perhaps it makes more sense to first compare groups
      comp = strcmp( ugdb->getUserNameS( fe1->userid() ),
                     ugdb->getUserNameS( fe2->userid() ) );
      if ( comp == 0 ) {
        comp = strcmp( ugdb->getGroupNameS( fe1->groupid() ),
                       ugdb->getGroupNameS( fe2->groupid() ) );
        if ( comp == 0 )
          comp = strcmp( fe1->name, fe2->name );
      }
      break;
    case SORT_INODE:
      if ( (loff_t)fe1->inode() < (loff_t)fe2->inode() ) comp = -1;
      else if ( (loff_t)fe1->inode() > (loff_t)fe2->inode() ) comp = 1;
      else comp = strcmp( fe1->name, fe2->name );
      break;
    case SORT_NLINK:
      if ( (loff_t)fe1->nlink() < (loff_t)fe2->nlink() ) comp = -1;
      else if ( (loff_t)fe1->nlink() > (loff_t)fe2->nlink() ) comp = 1;
      else comp = strcmp( fe1->name, fe2->name );
      break;
  }
  if((mode&SORT_REVERSE)==SORT_REVERSE) comp*=-1;
  if(dirm!=SORT_DIRMIXED) {
    if(fe1->isDir()==true) {
      if(fe2->isDir()==false) comp=(dirm==SORT_DIRLAST)?1:-1;
    } else {
      if(fe2->isDir()==true) comp=(dirm==SORT_DIRLAST)?-1:1;
    }
  }
  return comp;
}

void Verzeichnis::closeDir()
{
  if(files!=NULL) {
    FileEntry *fe=(FileEntry*)files->getFirstElement();
    while(fe!=NULL) {
      delete fe;
      fe=(FileEntry*)files->getNextElement();
    }
    delete files;
    files=NULL;
    _freesafe(actdir);
    actdir=NULL;
  }
  if(invalid_entries!=NULL) {
    FileEntry *fe=(FileEntry*)invalid_entries->getFirstElement();
    while(fe!=NULL) {
      delete fe;
      fe=(FileEntry*)invalid_entries->getNextElement();
    }
    delete invalid_entries;
    invalid_entries=NULL;
  }
}

int Verzeichnis::getSize()
{
  if(files!=NULL) return files->size();
  else return -1;
}

char *Verzeichnis::getDir()
{
  return actdir;
}

int Verzeichnis::sort( int mode )
{
    files->sort( sortfunction, mode );
    
    buildID2Pos();
    return 0;
}

int Verzeichnis::getUseSize()
{
  int size=0;
  if(files!=NULL) {
    int id=files->initEnum();
    FileEntry *fe=(FileEntry*)files->getFirstElement(id);
    while(fe!=NULL) {
      if(fe->use==true) size++;
      fe=(FileEntry*)files->getNextElement(id);
    }
    files->closeEnum(id);
  }
  return size;
}

ArrayList *Verzeichnis::getInvalidEntries()
{
  return invalid_entries;
}

void Verzeichnis::buildID2Pos()
{
    if ( files == NULL ||
         _highest_ID_plus_one < 1 ) {
        _id2pos.clear();
    } else {
        // some inital resizing but I don't rely on ID always
        // lower than size
        _id2pos.resize( _highest_ID_plus_one );
        std::fill( _id2pos.begin(),
                   _id2pos.end(),
                   -1 );
        
        int id = files->initEnum();
        FileEntry *fe = (FileEntry*)files->getFirstElement( id );
        int pos = 0;
        while ( fe != NULL ) {
            if ( fe->getID() < 0 ||
                 _id2pos[ fe->getID() ] != -1 ) {
                fprintf( stderr, "Worker error: file entry got wrong ID!" );
                exit( EXIT_FAILURE );
            } else {
                if ( fe->getID() >= static_cast<int>( _id2pos.size() ) ) {
                    fprintf( stderr, "Worker error: impossible ID detected!" );
                    exit( EXIT_FAILURE );
                }
                
                _id2pos[ fe->getID() ] = pos;
            }
            fe = (FileEntry*)files->getNextElement( id );
            pos++;
        }
        files->closeEnum( id );
    }
}

FileEntry *Verzeichnis::getEntryByID( int ID )
{
    if ( ID < 0 || ID > static_cast<int>( _id2pos.size() ) )
        return NULL;

    int pos = _id2pos[ID];
    if ( pos < 0 )
        return NULL;

    FileEntry *fe = (FileEntry*)files->getElementAt( pos );

    if ( fe->getID() != ID ) {
        // oops, wrong position, looks like someone modified the list
        fprintf( stderr, "Worker error: wrong position for entry!\n" );
        abort();
    }
    
    return fe;
}

void Verzeichnis::createIDs()
{
    if ( files == NULL )
        return;
    
    int id = files->initEnum();
    FileEntry *fe = (FileEntry*)files->getFirstElement( id );
    int ID = 0;
    while ( fe != NULL ) {
        fe->setID( ID++ );
        fe = (FileEntry*)files->getNextElement( id );
    }
    files->closeEnum( id );

    _highest_ID_plus_one = ID;

    _id_serial++;

    // -1 is not a valid ID
    if ( _id_serial == -1 )
        _id_serial++;
}

void Verzeichnis::removeElement( const FileEntry *fe )
{
    files->removeElement( fe );
    buildID2Pos();
    //TODO dirty flag benutzen, damit ich nur bei bedarf die Liste neu baue
    //TODO oder gleich map<ID,FE> benutzen, dann ist zwar der Zugriff langsamer, aber ich brauche keine ID2Pos mehr
}

int Verzeichnis::getSerialOfIDs()
{
    return _id_serial;
}

int Verzeichnis::getInvalidSerial()
{
    return -1;
}
