/* verzeichnis.hh
 * This file belongs to Worker, a filemanager for UNIX/X11.
 * Copyright (C) 2001-2007 Ralf Hoffmann.
 * You can contact me at: ralf@boomerangsworld.de
 *   or http://www.boomerangsworld.de/worker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef VERZEICHNIS_HH
#define VERZEICHNIS_HH

#include "wdefines.h"
#include "simplelist.hh"
#include <vector>

/* vvv Nur eins davon kann gleichzeitig aktiv sein, also durch numerieren */
enum { SORT_NAME = 0,
       SORT_SIZE,
       SORT_ACCTIME,
       SORT_MODTIME,
       SORT_CHGTIME,
       SORT_TYPE,
       SORT_OWNER,
       SORT_INODE,
       SORT_NLINK };

/* for checking valid sortmode */
#define ISVALID_SORTMODE(sortmode) ( ( ( sortmode & 0xff ) >= SORT_NAME ) && ( ( sortmode & 0xff ) <= SORT_NLINK ) )

/* vvv Dient zur Ver-Oderung mit dem Mode */
#define SORT_REVERSE 256
enum {SORT_DIRLAST=512,SORT_DIRMIXED=1024};

class Verzeichnis
{
public:
  Verzeichnis();
  ~Verzeichnis();
  Verzeichnis( const Verzeichnis &other );
  Verzeichnis &operator=( const Verzeichnis &other );

  ArrayList *getFiles();
  int readDir( const char *, bool dontWarnInvalid = false );
  void closeDir();
  int sort(int);
  int sort(int,bool force);
  int getSize();
  char *getDir();

  int getUseSize();
  static int sortfunction(void*,void*,int);
  ArrayList *getInvalidEntries();

    class FileEntry *getEntryByID( int ID );
    void removeElement( const class FileEntry *fe );
    int getSerialOfIDs();
    static int getInvalidSerial();
protected:

    ArrayList *files;
    char *actdir;
    
    ArrayList *invalid_entries;
    
    std::vector<int> _id2pos;
    void buildID2Pos();
    void createIDs();

    int _highest_ID_plus_one;
    int _id_serial;
};

#endif

